from typing import Dict, List

from ..utils.record import MetricMeter


class BaseMetric:
    def __init__(self, metric_names, primary_metric, **kwargs):
        assert primary_metric in metric_names
        self.metric_names = metric_names
        self.primary_metric = primary_metric
        self.metric_meter = {metric_name: MetricMeter(metric_name, ':6.2f') for metric_name in metric_names}

    def calibrate(self, pred, target, **kwargs):
        pass



from typing import Callable, Dict, Type

from .augmentation.transforms import create_transform
from .base import BaseCustomDataset, BaseHFDataset, BaseSampleLoader
from .classification import (
    ClassficationSampleLoader,
    ClassificationCustomDataset,
    ClassificationHFDataset,
)
from .detection import DetectionCustomDataset, DetectionSampleLoader
from .pose_estimation import (
    PoseEstimationCustomDataset,
    PoseEstimationSampleLoader,
)
from .segmentation import (
    SegmentationCustomDataset,
    SegmentationHFDataset,
    SegmentationSampleLoader,
)

CREATE_TRANSFORM = create_transform

CUSTOM_DATASET: Dict[str, Type[BaseCustomDataset]] = {
    'classification': ClassificationCustomDataset,
    'segmentation': SegmentationCustomDataset,
    'detection': DetectionCustomDataset,
    'pose_estimation': PoseEstimationCustomDataset,
}

HUGGINGFACE_DATASET: Dict[str, Type[BaseHFDataset]] = {
    'classification': ClassificationHFDataset,
    'segmentation': SegmentationHFDataset
}

DATA_SAMPLER: Dict[str, Type[BaseSampleLoader]] = {
    'classification': ClassficationSampleLoader,
    'segmentation': SegmentationSampleLoader,
    'detection': DetectionSampleLoader,
    'pose_estimation': PoseEstimationSampleLoader,
}



from typing import Dict, Type

from .classification import ClassificationPostprocessor
from .detection import DetectionPostprocessor
from .pose_estimation import PoseEstimationPostprocessor
from .segmentation import SegmentationPostprocessor

POSTPROCESSOR_DICT = {
    'fc': ClassificationPostprocessor,
    'all_mlp_decoder': SegmentationPostprocessor,
    'anchor_free_decoupled_head': DetectionPostprocessor,
    'pidnet': SegmentationPostprocessor,
    'anchor_decoupled_head': DetectionPostprocessor,
    'rtmcc': PoseEstimationPostprocessor,
}



from typing import Callable, Dict, Literal, Type

from .base import BaseMetric
from .classification import ClassificationMetric
from .detection import DetectionMetric
from .pose_estimation import PoseEstimationMetric
from .segmentation import SegmentationMetric

TASK_METRIC: Dict[Literal['classification', 'segmentation', 'detection'], Type[BaseMetric]] = {
    'classification': ClassificationMetric,
    'segmentation': SegmentationMetric,
    'detection': DetectionMetric,
    'pose_estimation': PoseEstimationMetric,
}

PHASE_LIST = ['train', 'valid', 'test']



from pathlib import Path
from typing import Callable, Dict, List, Type

import torch.nn as nn

from .backbones import cspdarknet, efficientformer, mixnet, mixtransformer, mobilenetv3, mobilevit, resnet, vit
from .base import ClassificationModel, DetectionModel, PoseEstimationModel, SegmentationModel, TaskModel
from .full import pidnet
from .heads.classification import fc
from .heads.detection import anchor_decoupled_head, anchor_free_decoupled_head
from .heads.pose_estimation import rtmcc
from .heads.segmentation import all_mlp_decoder
from .necks import fpn, yolopafpn

MODEL_BACKBONE_DICT: Dict[str, Callable[..., nn.Module]] = {
    'resnet': resnet,
    'mobilenetv3': mobilenetv3,
    'mixtransformer': mixtransformer,
    'mobilevit': mobilevit,
    'vit': vit,
    'efficientformer': efficientformer,
    'cspdarknet': cspdarknet,
    'mixnet': mixnet,
}

MODEL_NECK_DICT: Dict[str, Callable[..., nn.Module]] = {
    'fpn': fpn,
    'yolopafpn': yolopafpn,
}

MODEL_HEAD_DICT: Dict[str, Callable[..., nn.Module]] = {
    'classification': {
        'fc': fc,
    },
    'segmentation': {
        'all_mlp_decoder': all_mlp_decoder,
    },
    'detection': {
        'anchor_free_decoupled_head': anchor_free_decoupled_head,
        'anchor_decoupled_head': anchor_decoupled_head,
    },
    'pose_estimation': {
        'rtmcc': rtmcc,
    }
}

MODEL_FULL_DICT = {
    'pidnet': pidnet
}

SUPPORTING_MODEL_LIST = list(MODEL_BACKBONE_DICT.keys()) + list(MODEL_FULL_DICT.keys())
SUPPORTING_TASK_LIST: List[str] = ['classification', 'segmentation', 'detection', 'pose_estimation']

TASK_MODEL_DICT: Dict[str, Type[TaskModel]] = {
    'classification': ClassificationModel,
    'segmentation': SegmentationModel,
    'detection': DetectionModel,
    'pose_estimation': PoseEstimationModel,
}



from typing import Dict, List, Type

from .base import BasePipeline
from .evaluation import EvaluationPipeline
from .inference import InferencePipeline
from .task_processors.base import BaseTaskProcessor
from .task_processors.classification import ClassificationProcessor
from .task_processors.detection import DetectionProcessor
from .task_processors.pose_estimation import PoseEstimationProcessor
from .task_processors.segmentation import SegmentationProcessor
from .train import TrainingPipeline

# TODO: Temporary defined. It should be integrated with `..models.registry.SUPPORTING_TASK_LIST`
SUPPORTING_TASK_LIST: List[str] = ['classification', 'segmentation', 'detection', 'pose_estimation']

TASK_PROCESSOR: Dict[str, Type[BaseTaskProcessor]] = {
    'classification': ClassificationProcessor,
    'segmentation': SegmentationProcessor,
    'detection': DetectionProcessor,
    'pose_estimation': PoseEstimationProcessor,
}

PIPELINES: Dict[str, Type[BasePipeline]] = {
    'train': TrainingPipeline,
    'evaluation': EvaluationPipeline,
    'inference': InferencePipeline,
}



from .registry import POSTPROCESSOR_DICT


def build_postprocessor(task: str, conf_model):
    head_name = conf_model.architecture.full.name if conf_model.single_task_model else conf_model.architecture.head.name
    if head_name not in POSTPROCESSOR_DICT:
        return None
    return POSTPROCESSOR_DICT[head_name](conf_model)



import os
from abc import abstractmethod
from typing import Callable, Optional, Tuple, Union

import torch
import torch.nn as nn
from loguru import logger
from omegaconf import OmegaConf

from .utils import BackboneOutput, DetectionModelOutput, ModelOutput


class TaskModel(nn.Module):
    def __init__(self, conf_model, backbone, neck, head, freeze_backbone: bool = False) -> None:
        super(TaskModel, self).__init__()
        self.task = conf_model.task
        self.name = conf_model.name
        self.backbone_name = conf_model.architecture.backbone.name
        if neck:
            self.neck_name = conf_model.architecture.neck.name
        self.head_name = conf_model.architecture.head.name

        self.backbone = backbone
        if neck:
            self.neck = neck
        self.head = head

        if freeze_backbone:
            self._freeze_backbone()
            logger.info(f"Freeze! {self.backbone_name} is now freezed. Now only tuning with {self.head_name}.")

    def _freeze_backbone(self):
        for m in self.backbone.parameters():
            m.requires_grad = False

    @property
    def head_list(self):
        return ('head')

    @property
    def device(self):
        return next(self.parameters()).device

    def _get_name(self):
        if hasattr(self, 'neck'):
            return f"{self.__class__.__name__}[task={self.task}, backbone={self.backbone_name}, neck={self.neck_name}, head={self.head_name}]"
        else:
            return f"{self.__class__.__name__}[task={self.task}, backbone={self.backbone_name}, head={self.head_name}]"

    @abstractmethod
    def forward(self, x, label_size=None, targets=None):
        raise NotImplementedError


class ClassificationModel(TaskModel):
    def __init__(self, conf_model, backbone, neck, head, freeze_backbone=False) -> None:
        super().__init__(conf_model, backbone, neck, head, freeze_backbone)

    def forward(self, x, label_size=None, targets=None):
        features: BackboneOutput = self.backbone(x)
        out: ModelOutput = self.head(features['last_feature'])
        return out


class SegmentationModel(TaskModel):
    def __init__(self, conf_model, backbone, neck, head, freeze_backbone=False) -> None:
        super().__init__(conf_model, backbone, neck, head, freeze_backbone)

    def forward(self, x, label_size=None, targets=None):
        features: BackboneOutput = self.backbone(x)
        if hasattr(self, 'neck'):
            features: BackboneOutput = self.neck(features['intermediate_features'])
        out: ModelOutput = self.head(features['intermediate_features'])
        return out


class DetectionModel(TaskModel):
    def __init__(self, conf_model, backbone, neck, head, freeze_backbone=False) -> None:
        super().__init__(conf_model, backbone, neck, head, freeze_backbone)

    def forward(self, x, label_size=None, targets=None):
        features: BackboneOutput = self.backbone(x)
        if hasattr(self, 'neck'):
            features: BackboneOutput = self.neck(features['intermediate_features'])
        out: DetectionModelOutput = self.head(features['intermediate_features'])
        return out

class PoseEstimationModel(TaskModel):
    def __init__(self, conf_model, backbone, neck, head, freeze_backbone=False) -> None:
        super().__init__(conf_model, backbone, neck, head, freeze_backbone)

    def forward(self, x, label_size=None, targets=None):
        features: BackboneOutput = self.backbone(x)
        if hasattr(self, 'neck'):
            features: BackboneOutput = self.neck(features['intermediate_features'])
        out: DetectionModelOutput = self.head(features['intermediate_features'])
        return out




from omegaconf import OmegaConf

from .registry import SCHEDULER_DICT


def build_scheduler(optimizer, training_conf):
    scheduler_conf = training_conf.scheduler
    scheduler_name = scheduler_conf.name
    num_epochs = training_conf.epochs

    assert scheduler_name in SCHEDULER_DICT, f"{scheduler_name} not in scheduler dict!"
    lr_scheduler = SCHEDULER_DICT[scheduler_name](optimizer, scheduler_conf, num_epochs)

    return lr_scheduler, num_epochs



from .builder import build_dataloader, build_dataset
from .registry import CREATE_TRANSFORM, CUSTOM_DATASET, DATA_SAMPLER, HUGGINGFACE_DATASET
from .utils.constants import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD



from .visualizer import ClassificationVisualizer, DetectionVisualizer, PoseEstimationVisualizer, SegmentationVisualizer

VISUALIZER = {
    'classification': ClassificationVisualizer,
    'segmentation': SegmentationVisualizer,
    'detection': DetectionVisualizer,
    'pose_estimation': PoseEstimationVisualizer,
}



import json
from dataclasses import asdict
from pathlib import Path
from statistics import mean
from typing import Dict, List, Literal, Optional, final

import torch
import torch.nn as nn
from loguru import logger
from omegaconf import DictConfig
from torch.utils.data import DataLoader
from tqdm import tqdm

from ..loggers.base import TrainingLogger
from ..losses.builder import LossFactory
from ..metrics.builder import MetricFactory
from ..utils.record import EvaluationSummary, Timer
from ..utils.stats import get_params_and_macs
from .base import BasePipeline
from .task_processors.base import BaseTaskProcessor

NUM_SAMPLES = 16


class EvaluationPipeline(BasePipeline):
    def __init__(
        self,
        conf: DictConfig,
        task: str,
        task_processor: BaseTaskProcessor,
        model_name: str,
        model: nn.Module,
        logger: Optional[TrainingLogger],
        timer: Timer,
        loss_factory: LossFactory,
        metric_factory: MetricFactory,
        eval_dataloader: DataLoader,
        single_gpu_or_rank_zero: bool,
    ):
        super(EvaluationPipeline, self).__init__(conf, task, task_processor, model_name, model, logger, timer)
        self.loss_factory = loss_factory
        self.metric_factory = metric_factory
        self.eval_dataloader = eval_dataloader
        self.single_gpu_or_rank_zero = single_gpu_or_rank_zero

    @final
    def _is_ready(self):
        assert self.model is not None, "`self.model` is not defined!"
        return True

    @property
    def valid_loss(self):
        return self.loss_factory.result('valid').get('total').avg

    @torch.no_grad()
    def evaluation(self, num_samples=NUM_SAMPLES):
        self._is_ready()
        self.timer.start_record(name='evaluation')

        num_returning_samples = 0
        returning_samples = []
        outputs = []
        for _idx, batch in enumerate(tqdm(self.eval_dataloader, leave=False)):
            out = self.task_processor.valid_step(self.model, batch, self.loss_factory, self.metric_factory)
            if out is not None:
                outputs.append(out)
                if num_returning_samples < num_samples:
                    returning_samples.append(out)
                    num_returning_samples += len(out['pred'])
        self.task_processor.get_metric_with_all_outputs(outputs, phase='valid', metric_factory=self.metric_factory)

        self.timer.end_record(name='evaluation')
        if self.single_gpu_or_rank_zero:
            time_for_evaluation = self.timer.get(name='evaluation', as_pop=False)
            self.log_end_evaluation(time_for_evaluation=time_for_evaluation, valid_samples=returning_samples)

    def log_end_evaluation(
        self,
        time_for_evaluation: float,
        valid_samples: Optional[List] = None,
    ):
        losses = self.loss_factory.result('valid')
        metrics = self.metric_factory.result('valid')
        self.log_results(
            prefix='evaluation',
            samples=valid_samples,
            losses=losses,
            metrics=metrics,
            elapsed_time=time_for_evaluation,
        )
        self.save_summary(losses, metrics, time_for_evaluation)

    def save_summary(self, losses, metrics, time_for_evaluation):
        macs, params = get_params_and_macs(self.model, self.sample_input.float())
        evaluation_summary = EvaluationSummary(
            losses=losses,
            metrics=metrics,
            metrics_list=self.metric_factory.metric_names,
            primary_metric=self.metric_factory.primary_metric,
            macs=macs,
            params=params,
            total_evaluation_time=time_for_evaluation,
            success=True,
        )

        logger.info(f"[Model stats] Params: {(params/1e6):.2f}M | MACs: {(macs/1e9):.2f}G")
        logging_dir = self.logger.result_dir
        summary_path = Path(logging_dir) / "evaluation_summary.json"

        with open(summary_path, 'w') as f:
            json.dump(asdict(evaluation_summary), f, indent=4)
        logger.info(f"Model evaluation summary saved at {str(summary_path)}")



from typing import Union

from omegaconf import DictConfig
import torch
import torch.nn as nn
from torch import Tensor
from torch.fx.proxy import Proxy

from ....utils import ModelOutput
from ....op.registry import ACTIVATION_REGISTRY


class FC(nn.Module):
    def __init__(self, feature_dim: int, num_classes: int, params: DictConfig) -> None:
        super(FC, self).__init__()
        dropout_prob = params.dropout_prob
        num_layers = params.num_layers

        assert num_layers >= 1, "num_hidden_layers must be integer larger than 0"

        prev_size = feature_dim
        classifier = []
        for _ in range(num_layers - 1):
            classifier.append(nn.Linear(prev_size, params.intermediate_channels))
            classifier.append(ACTIVATION_REGISTRY[params.act_type]())
            prev_size = params.intermediate_channels
        classifier.append(nn.Dropout(p=dropout_prob))
        classifier.append(nn.Linear(prev_size, num_classes))
        self.classifier = nn.Sequential(*classifier)
        
    def forward(self, x: Union[Tensor, Proxy]) -> ModelOutput:
        x = self.classifier(x)
        return ModelOutput(pred=x)
    
def fc(feature_dim, num_classes, conf_model_head, **kwargs) -> FC:
    return FC(feature_dim=feature_dim, num_classes=num_classes, params=conf_model_head.params)


from abc import ABC, abstractmethod
from typing import Dict, List, Literal, Optional

import torch
import torch.nn as nn
from omegaconf import DictConfig

from ..loggers.base import TrainingLogger
from ..utils.record import Timer
from .task_processors.base import BaseTaskProcessor


class BasePipeline(ABC):
    def __init__(
        self,
        conf: DictConfig,
        task: str,
        task_processor: BaseTaskProcessor,
        model_name: str,
        model: nn.Module,
        logger: Optional[TrainingLogger],
        timer: Timer,
    ):
        super(BasePipeline, self).__init__()
        self.conf = conf
        self.task = task
        self.task_processor = task_processor
        self.model_name = model_name
        self.save_dtype = next(model.parameters()).dtype
        self.model = model.float()
        self.logger = logger
        self.timer = timer

    @property
    def sample_input(self):
        return torch.randn((1, 3, self.conf.logging.onnx_input_size[0], self.conf.logging.onnx_input_size[1]))

    def log_results(
        self,
        prefix: Literal['training', 'validation', 'evaluation', 'inference'],
        epoch: Optional[int] = None,
        samples: Optional[List] = None,
        losses : Optional[Dict] = None,
        metrics: Optional[Dict] = None,
        learning_rate: Optional[float] = None,
        elapsed_time: Optional[float] = None,
    ):
        self.logger.log(
            prefix=prefix,
            epoch=epoch,
            samples=samples,
            losses=losses,
            metrics=metrics,
            learning_rate=learning_rate,
            elapsed_time=elapsed_time
        )

    @abstractmethod
    def save_summary(self):
        raise NotImplementedError



from typing import Literal

import torch

from .base import BaseTaskProcessor


class SegmentationProcessor(BaseTaskProcessor):
    def __init__(self, conf, postprocessor, devices, **kwargs):
        super(SegmentationProcessor, self).__init__(conf, postprocessor, devices, **kwargs)

    def train_step(self, train_model, batch, optimizer, loss_factory, metric_factory):
        train_model.train()
        batch['indices']
        images = batch['pixel_values'].to(self.devices)
        labels = batch['labels'].long().to(self.devices)
        target = {'target': labels}

        if 'edges' in batch:
            bd_gt = batch['edges']
            target['bd_gt'] = bd_gt.to(self.devices)

        optimizer.zero_grad()

        with torch.cuda.amp.autocast(enabled=self.mixed_precision):
            out = train_model(images)
            loss_factory.calc(out, target, phase='train')

        loss_factory.backward(self.grad_scaler)
        self.grad_scaler.step(optimizer)
        self.grad_scaler.update()

        out = {k: v.detach() for k, v in out.items()}
        pred = self.postprocessor(out)

        labels = labels.detach().cpu().numpy() # Change it to numpy before compute metric
        if self.conf.distributed:
            gathered_pred = [None for _ in range(torch.distributed.get_world_size())]
            gathered_labels = [None for _ in range(torch.distributed.get_world_size())]

            torch.distributed.gather_object(pred, gathered_pred if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.gather_object(labels, gathered_labels if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.barrier()
            if torch.distributed.get_rank() == 0:
                [metric_factory.update(g_pred, g_labels, phase='train') for g_pred, g_labels in zip(gathered_pred, gathered_labels)]
        else:
            metric_factory.update(pred, labels, phase='train')

    def valid_step(self, eval_model, batch, loss_factory, metric_factory):
        eval_model.eval()
        indices = batch['indices']
        images = batch['pixel_values'].to(self.devices)
        labels = batch['labels'].long().to(self.devices)
        target = {'target': labels}

        if 'edges' in batch:
            bd_gt = batch['edges']
            target['bd_gt'] = bd_gt.to(self.devices)

        out = eval_model(images)
        loss_factory.calc(out, target, phase='valid')

        pred = self.postprocessor(out)

        indices = indices.numpy()
        labels = labels.detach().cpu().numpy() # Change it to numpy before compute metric
        if self.conf.distributed:
            gathered_pred = [None for _ in range(torch.distributed.get_world_size())]
            gathered_labels = [None for _ in range(torch.distributed.get_world_size())]

            # Remove dummy samples, they only come in distributed environment
            pred = pred[indices != -1]
            labels = labels[indices != -1]
            torch.distributed.gather_object(pred, gathered_pred if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.gather_object(labels, gathered_labels if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.barrier()
            if torch.distributed.get_rank() == 0:
                [metric_factory.update(g_pred, g_labels, phase='valid') for g_pred, g_labels in zip(gathered_pred, gathered_labels)]
        else:
            metric_factory.update(pred, labels, phase='valid')

        logs = {
            'images': images.detach().cpu().numpy(),
            'target': labels,
            'pred': pred
        }
        if 'edges' in batch:
            logs.update({
                'bd_gt': bd_gt.detach().cpu().numpy()
            })
        return dict(logs.items())

    def test_step(self, test_model, batch):
        test_model.eval()
        indices = batch['indices']
        images = batch['pixel_values']
        images = images.to(self.devices)

        out = test_model(images)

        pred = self.postprocessor(out)

        indices = indices.numpy()
        if self.conf.distributed:
            gathered_pred = [None for _ in range(torch.distributed.get_world_size())]

            # Remove dummy samples, they only come in distributed environment
            pred = pred[indices != -1]
            torch.distributed.gather_object(pred, gathered_pred if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.barrier()
            if torch.distributed.get_rank() == 0:
                gathered_pred = sum(gathered_pred, [])
                pred = gathered_pred

        if self.single_gpu_or_rank_zero:
            results = {
                'images': images.detach().cpu().numpy(),
                'pred': pred
            }
            return results

    def get_metric_with_all_outputs(self, outputs, phase: Literal['train', 'valid'], metric_factory):
        pass



from typing import Callable, Dict

from .custom.image_proc import (
    AutoAugment,
    CenterCrop,
    ColorJitter,
    HSVJitter,
    Pad,
    PoseTopDownAffine,
    RandomCrop,
    RandomErasing,
    RandomHorizontalFlip,
    RandomResize,
    RandomResizedCrop,
    RandomVerticalFlip,
    Resize,
    TrivialAugmentWide,
)
from .custom.mixing import Mixing
from .custom.mosaic import MosaicDetection

TRANSFORM_DICT: Dict[str, Callable] = {
    'centercrop': CenterCrop,
    'colorjitter': ColorJitter,
    'pad': Pad,
    'randomcrop': RandomCrop,
    'randomresizedcrop': RandomResizedCrop,
    'randomhorizontalflip': RandomHorizontalFlip,
    'randomresize': RandomResize,
    'randomverticalflip': RandomVerticalFlip,
    'randomerasing': RandomErasing,
    'resize': Resize,
    'mixing': Mixing,
    'mosaicdetection': MosaicDetection,
    'trivialaugmentwide': TrivialAugmentWide,
    'autoaugment': AutoAugment,
    'hsvjitter': HSVJitter,
    'posetopdownaffine': PoseTopDownAffine,
}



from typing import Literal

import torch.nn as nn
from omegaconf import DictConfig

from .registry import OPTIMIZER_DICT


def build_optimizer(
    model_or_params,
    optimizer_conf: DictConfig,
):
    parameters = model_or_params.parameters() if isinstance(model_or_params, nn.Module) else model_or_params

    opt_name: Literal['sgd', 'adam', 'adamw', 'adamax', 'adadelta', 'adagrad', 'rmsprop'] = optimizer_conf.name.lower()
    assert opt_name in OPTIMIZER_DICT

    optimizer = OPTIMIZER_DICT[opt_name](parameters, optimizer_conf)
    return optimizer



import json
from dataclasses import asdict
from pathlib import Path
from statistics import mean
from typing import Dict, List, Literal, Optional, final

import torch
import torch.nn as nn
from loguru import logger
from omegaconf import DictConfig
from torch.utils.data import DataLoader
from tqdm import tqdm

from ..loggers.base import TrainingLogger
from ..utils.record import InferenceSummary, Timer
from ..utils.stats import get_params_and_macs
from .base import BasePipeline
from .task_processors.base import BaseTaskProcessor

NUM_SAMPLES = 16


class InferencePipeline(BasePipeline):
    def __init__(
        self,
        conf: DictConfig,
        task: str,
        task_processor: BaseTaskProcessor,
        model_name: str,
        model: nn.Module,
        logger: Optional[TrainingLogger],
        timer: Timer,
        test_dataloader: DataLoader,
        single_gpu_or_rank_zero: bool,
    ):
        super(InferencePipeline, self).__init__(conf, task, task_processor, model_name, model, logger, timer)
        self.test_dataloader = test_dataloader
        self.single_gpu_or_rank_zero = single_gpu_or_rank_zero

    @final
    def _is_ready(self):
        assert self.model is not None, "`self.model` is not defined!"
        return True

    @torch.no_grad()
    def inference(self):
        self._is_ready()
        self.timer.start_record(name='inference')

        num_returning_samples = 0
        returning_samples = []
        outputs = []
        for _idx, batch in enumerate(tqdm(self.test_dataloader, leave=False)):
            out = self.task_processor.test_step(self.model, batch)
            if out is not None:
                outputs.append(out)
                if num_returning_samples < NUM_SAMPLES: # TODO: Save all output or set by config
                    returning_samples.append(out)
                    num_returning_samples += len(out['pred'])

        self.timer.end_record(name='inference')
        if self.single_gpu_or_rank_zero:
            time_for_inference = self.timer.get(name='inference', as_pop=False)
            self.log_end_inference(time_for_inference=time_for_inference, valid_samples=returning_samples)

    def log_end_inference(
        self,
        time_for_inference: float,
        valid_samples: Optional[List] = None,
    ):
        self.log_results(
            prefix='inference',
            samples=valid_samples,
            elapsed_time=time_for_inference,
        )
        self.save_summary(time_for_inference)

    def save_summary(self, time_for_inference):
        macs, params = get_params_and_macs(self.model, self.sample_input.float())
        inference_summary = InferenceSummary(
            macs=macs,
            params=params,
            total_inference_time=time_for_inference,
            success=True,
        )

        logger.info(f"[Model stats] Params: {(params/1e6):.2f}M | MACs: {(macs/1e9):.2f}G")
        logging_dir = self.logger.result_dir
        summary_path = Path(logging_dir) / "inference_summary.json"

        with open(summary_path, 'w') as f:
            json.dump(asdict(inference_summary), f, indent=4)
        logger.info(f"Model inference summary saved at {str(summary_path)}")



from typing import List

import numpy as np

from ..base import BaseMetric


class PoseEstimationMetric(BaseMetric):
    SUPPORT_METRICS: List[str] = ['pck']

    def __init__(self, **kwargs):
        # TODO: Select metrics by user
        metric_names: List[str] = ['pck']
        primary_metric: str = 'pck'

        assert set(metric_names).issubset(PoseEstimationMetric.SUPPORT_METRICS)
        super().__init__(metric_names=metric_names, primary_metric=primary_metric)
        # TODO: Get from config
        self.thr = 0.05
        self.input_size = (256, 256)

    def _calc_distances(self, preds, gts, mask, norm_factor):
        N, K, _ = preds.shape
        # set mask=0 when norm_factor==0
        _mask = mask.copy()
        _mask[np.where((norm_factor == 0).sum(1))[0], :] = False

        distances = np.full((N, K), -1, dtype=np.float32)
        # handle invalid values
        norm_factor[np.where(norm_factor <= 0)] = 1e6
        distances[_mask] = np.linalg.norm(((preds - gts) / norm_factor[:, None, :])[_mask], axis=-1)
        return distances.T

    def _distance_acc(self, distances, thr):
        distance_valid = distances != -1
        num_distance_valid = distance_valid.sum()
        if num_distance_valid > 0:
            return (distances[distance_valid] < thr).sum() / num_distance_valid
        return -1

    def keypoint_pck_accuracy(self, pred, gt, mask, thr, norm_factor):
        distances = self._calc_distances(pred, gt, mask, norm_factor)
        acc = np.array([self._distance_acc(d, thr) for d in distances])
        valid_acc = acc[acc >= 0]
        cnt = len(valid_acc)
        avg_acc = valid_acc.mean() if cnt > 0 else 0.0
        # ``acc`` contains accuracy of each keypoint
        return acc, avg_acc, cnt

    def calibrate(self, pred, target, **kwargs):
        N, _, _ = pred.shape

        mask = target[..., 2] > 0
        target = target[..., :2]

        # if normalize is None:
        #     normalize = np.tile(np.array([[H, W]]), (N, 1))
        # Normalize by box
        normalize = np.tile(np.array([[self.input_size[0], self.input_size[1]]]), (N, 1))

        acc, avg_acc, cnt = self.keypoint_pck_accuracy(pred, target, mask, self.thr, normalize)
        self.metric_meter['pck'].update(avg_acc)



from pathlib import Path
from typing import Dict, List, Literal, Optional, Tuple, Union

from .base import TrainingLogger


def build_logger(conf, task: str, model_name: str, step_per_epoch: int, class_map: Dict[int, str], num_sample_images: int, result_dir: Union[Path, str], epoch: Optional[int] = None):
    training_logger = TrainingLogger(conf,
                                     task=task, model=model_name,
                                     step_per_epoch=step_per_epoch,
                                     class_map=class_map, num_sample_images=num_sample_images,
                                     result_dir=result_dir,
                                     epoch=epoch)

    return training_logger



from typing import Dict, Type

from torch.optim.optimizer import Optimizer

from .custom import SGD, Adadelta, Adagrad, Adam, Adamax, AdamW, RMSprop

OPTIMIZER_DICT: Dict[str, Type[Optimizer]] = {
    'adamw': AdamW,
    'adam': Adam,
    'adadelta': Adadelta,
    'adagrad': Adagrad,
    'rmsprop': RMSprop,
    'adamax': Adamax,
    'sgd': SGD,
}



from typing import Optional

from ..models.utils import ModelOutput

TOPK_MAX = 20


class ClassificationPostprocessor():
    def __init__(self, conf_model):
        pass

    def __call__(self, outputs: ModelOutput, k: Optional[int]=None):
        pred = outputs['pred']
        maxk = min(TOPK_MAX, pred.size()[1])
        if k:
            maxk = min(k, maxk)
        _, pred = pred.topk(maxk, 1, True, True)
        return pred.detach().cpu().numpy()
